from datetime import datetime, timezone
from time import time
from flask import current_app, g
from flask.sessions import SecureCookieSessionInterface
from flask_ldap3_login import AuthenticationResponseStatus
from flask_login import login_user, logout_user
from itsdangerous import (TimedJSONWebSignatureSerializer
                          as Serializer, BadSignature, SignatureExpired)
from opsy.flask_extensions import ldap_manager


class APISessionInterface(SecureCookieSessionInterface):
    """Prevent creating session from API requests."""

    def save_session(self, *args, **kwargs):
        if g.get('login_via_header'):
            return
        super(APISessionInterface, self).save_session(*args, **kwargs)


def login(user_name, password, force_renew=False):
    # First we try local login
    user = local_login(user_name, password)
    # If that failed we try ldap
    if not user:
        user = ldap_login(user_name, password)
    # If that failed we bail out
    if not user:
        current_app.logger.info(
            f'All login methods failed for user {user_name}.')
        return False
    if not login_user(user):
        current_app.logger.info(f'Login for user {user_name} failed. Disabled?')
        return False
    current_app.logger.info(f'User {user_name} logged in successfully.')
    # Force renewal of the token if it was requested.
    return create_token(user, force_renew=force_renew)


def local_login(user_name, password):
    from opsy.auth.models import User
    current_app.logger.info(f'Attempting local login for user {user_name}...')
    user = User.query.filter_by(name=user_name).first()
    if not user:
        current_app.logger.info(
            f'User {user_name} does not exist, skipping local login.')
        return False
    if user.ldap_user:
        current_app.logger.info(
            f'User {user_name} is an LDAP user, skipping local login.')
        return False
    if not user.verify_password(password):
        current_app.logger.info(f'Incorrect password for user {user_name}.')
        return False
    return user


def ldap_login(user_name, password):  # pylint: disable=too-many-branches
    from opsy.auth.models import User, Role
    current_app.logger.info(f'Attempting LDAP login for user {user_name}...')
    if not current_app.config.opsy['auth']['ldap_enabled']:
        current_app.logger.info(
            f'LDAP login disabled, unable to login user {user_name}.')
        return False
    user = User.query.filter_by(name=user_name).first()
    if user and not user.ldap_user:
        current_app.logger.info(
            f'User {user_name} is a local user, skipping LDAP login.')
        return False
    result = ldap_manager.authenticate(user_name, password)
    if not result.status == AuthenticationResponseStatus.success:
        current_app.logger.info(
            f'LDAP login failed for user {user_name}. Incorrect password?')
        return False
    full_name_attr = \
        current_app.config.opsy['auth']['ldap_user_full_name_attr']
    email_attr = current_app.config.opsy['auth']['ldap_user_email_attr']
    group_name_attr = \
        current_app.config.opsy['auth']['ldap_group_name_attr']
    if isinstance(result.user_info[email_attr], list):
        email = result.user_info[email_attr][0]
    else:
        email = result.user_info[email_attr]
    full_name = result.user_info[full_name_attr]
    user = User.query.filter_by(name=result.user_id).first()
    if not user:
        user = User.create(
            user_name, email=email, full_name=full_name, ldap_user=True)
    else:
        user.update(email=email, full_name=full_name)
    groups = []
    for group in result.user_groups:
        if isinstance(group[group_name_attr], list):
            groups.append(group[group_name_attr][0])
        else:
            groups.append(group[group_name_attr])
    for role in user.roles:
        if role.ldap_group:
            user.roles.remove(role)
    for role in Role.query.filter(Role.ldap_group.in_(groups)).all():
        user.roles.append(role)
    return user.save()


def logout(user):
    if not user:
        return
    username = user.name
    user.session_token = None
    user.session_token_expires_at = None
    for role in user.roles:
        if role.ldap_group:
            user.roles.remove(role)
    user.save()
    logout_user()
    current_app.logger.info(f'{username} logged out successfully.')


def create_token(user, force_renew=False):
    if not force_renew and verify_token(user) is not None:
        return user
    ttl = current_app.config.opsy['auth']['session_token_ttl']
    seri = Serializer(
        current_app.config['SECRET_KEY'], expires_in=ttl)
    user.session_token = seri.dumps({'id': user.id}).decode('ascii')
    _, header = seri.loads(user.session_token, return_header=True)
    user.session_token_expires_at = datetime.fromtimestamp(
        header.get('exp'), tz=timezone.utc)
    return user.save()


def verify_token(user):  # pylint: disable=R0911
    if user is None:
        return None
    ttl = current_app.config.opsy['auth']['session_token_ttl']
    seri = Serializer(current_app.config['SECRET_KEY'])
    try:
        data, header = seri.loads(user.session_token, return_header=True)
    except TypeError:
        return None  # we don't currently have a token
    except SignatureExpired:
        current_app.logger.info(f'Expired token for user {user.name}.')
        return None  # valid token, but expired
    except BadSignature:
        current_app.logger.info(f'Invalid token for user {user.name}.')
        return None  # invalid token
    if int(user.session_token_expires_at.replace(
            tzinfo=timezone.utc).timestamp()) > time() + ttl:
        current_app.logger.info(f'Invalid token ttl for user {user.name}.')
        return None  # ttl in config has been reduced
    # These next two shouldn't ever happen
    if data.get('id') != user.id:
        current_app.logger.info(f'Invalid token id for user {user.name}.')
        return None  # user id doesn't match token payload
    if user.session_token_expires_at != datetime.fromtimestamp(
            header.get('exp'), tz=timezone.utc):
        current_app.logger.info(f'Invalid token exp for user {user.name}.')
        return None  # expire times don't match
    return user


def load_user(session_token):
    from opsy.auth.models import User
    user = User.get_by_token(session_token)
    return verify_token(user)


def load_user_from_request(request):
    from opsy.auth.models import User
    session_token = request.headers.get('X-AUTH-TOKEN')
    user = verify_token(User.get_by_token(session_token))
    g.login_via_header = True
    return user



"""This is the initial schema for Opsy.

Revision ID: 22d14acc37df
Revises: 
Create Date: 2019-10-03 16:21:51.396397

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '22d14acc37df'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('roles',
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('id', sa.String(length=36), nullable=False),
    sa.Column('name', sa.String(length=128), nullable=False),
    sa.Column('ldap_group', sa.String(length=128), nullable=True),
    sa.Column('description', sa.Text(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_roles_ldap_group'), 'roles', ['ldap_group'], unique=False)
    op.create_index(op.f('ix_roles_name'), 'roles', ['name'], unique=True)
    op.create_table('users',
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('id', sa.String(length=36), nullable=False),
    sa.Column('name', sa.String(length=128), nullable=False),
    sa.Column('full_name', sa.String(length=64), nullable=True),
    sa.Column('email', sa.String(length=64), nullable=True),
    sa.Column('password_hash', sa.String(length=128), nullable=True),
    sa.Column('session_token', sa.String(length=255), nullable=True),
    sa.Column('session_token_expires_at', sa.DateTime(), nullable=True),
    sa.Column('enabled', sa.Boolean(), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('session_token', name='sess_uc')
    )
    op.create_index(op.f('ix_users_email'), 'users', ['email'], unique=False)
    op.create_index(op.f('ix_users_name'), 'users', ['name'], unique=True)
    op.create_index(op.f('ix_users_session_token'), 'users', ['session_token'], unique=False)
    op.create_table('zones',
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('id', sa.String(length=36), nullable=False),
    sa.Column('name', sa.String(length=128), nullable=False),
    sa.Column('vars', sa.JSON(), nullable=True),
    sa.Column('description', sa.Text(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_zones_name'), 'zones', ['name'], unique=True)
    op.create_table('groups',
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('id', sa.String(length=36), nullable=False),
    sa.Column('name', sa.String(length=128), nullable=False),
    sa.Column('parent_id', sa.String(length=36), nullable=True),
    sa.Column('zone_id', sa.String(length=36), nullable=True),
    sa.Column('default_priority', sa.BigInteger(), nullable=True),
    sa.Column('vars', sa.JSON(), nullable=True),
    sa.ForeignKeyConstraint(['parent_id'], ['groups.id'], ),
    sa.ForeignKeyConstraint(['zone_id'], ['zones.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('name', 'zone_id'),
    sa.UniqueConstraint('name', 'zone_id')
    )
    op.create_index('groups_name_zone_id_key_not_null', 'groups', ['name', 'zone_id'], unique=True, postgresql_where=sa.text('zone_id IS NOT NULL'), sqlite_where=sa.text('zone_id IS NOT NULL'))
    op.create_index('groups_name_zone_id_key_null', 'groups', ['name'], unique=True, postgresql_where=sa.text('zone_id IS NULL'), sqlite_where=sa.text('zone_id IS NULL'))
    op.create_index(op.f('ix_groups_name'), 'groups', ['name'], unique=False)
    op.create_index(op.f('ix_groups_parent_id'), 'groups', ['parent_id'], unique=False)
    op.create_index(op.f('ix_groups_zone_id'), 'groups', ['zone_id'], unique=False)
    op.create_table('hosts',
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('id', sa.String(length=36), nullable=False),
    sa.Column('name', sa.String(length=128), nullable=False),
    sa.Column('zone_id', sa.String(length=36), nullable=False),
    sa.Column('vars', sa.JSON(), nullable=True),
    sa.ForeignKeyConstraint(['zone_id'], ['zones.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_hosts_name'), 'hosts', ['name'], unique=True)
    op.create_index(op.f('ix_hosts_zone_id'), 'hosts', ['zone_id'], unique=False)
    op.create_table('permissions',
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('id', sa.String(length=36), nullable=False),
    sa.Column('role_id', sa.String(length=36), nullable=False),
    sa.Column('name', sa.String(length=128), nullable=False),
    sa.ForeignKeyConstraint(['role_id'], ['roles.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('role_id', 'name')
    )
    op.create_index(op.f('ix_permissions_name'), 'permissions', ['name'], unique=False)
    op.create_index(op.f('ix_permissions_role_id'), 'permissions', ['role_id'], unique=False)
    op.create_table('role_mappings',
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('id', sa.String(length=36), nullable=False),
    sa.Column('role_id', sa.String(length=36), nullable=False),
    sa.Column('user_id', sa.String(length=36), nullable=False),
    sa.ForeignKeyConstraint(['role_id'], ['roles.id'], ondelete='CASCADE'),
    sa.ForeignKeyConstraint(['user_id'], ['users.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_role_mappings_role_id'), 'role_mappings', ['role_id'], unique=False)
    op.create_index(op.f('ix_role_mappings_user_id'), 'role_mappings', ['user_id'], unique=False)
    op.create_table('host_group_mappings',
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('id', sa.String(length=36), nullable=False),
    sa.Column('host_id', sa.String(length=36), nullable=False),
    sa.Column('group_id', sa.String(length=36), nullable=False),
    sa.Column('priority', sa.BigInteger(), nullable=True),
    sa.ForeignKeyConstraint(['group_id'], ['groups.id'], ondelete='CASCADE'),
    sa.ForeignKeyConstraint(['host_id'], ['hosts.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_host_group_mappings_group_id'), 'host_group_mappings', ['group_id'], unique=False)
    op.create_index(op.f('ix_host_group_mappings_host_id'), 'host_group_mappings', ['host_id'], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f('ix_host_group_mappings_host_id'), table_name='host_group_mappings')
    op.drop_index(op.f('ix_host_group_mappings_group_id'), table_name='host_group_mappings')
    op.drop_table('host_group_mappings')
    op.drop_index(op.f('ix_role_mappings_user_id'), table_name='role_mappings')
    op.drop_index(op.f('ix_role_mappings_role_id'), table_name='role_mappings')
    op.drop_table('role_mappings')
    op.drop_index(op.f('ix_permissions_role_id'), table_name='permissions')
    op.drop_index(op.f('ix_permissions_name'), table_name='permissions')
    op.drop_table('permissions')
    op.drop_index(op.f('ix_hosts_zone_id'), table_name='hosts')
    op.drop_index(op.f('ix_hosts_name'), table_name='hosts')
    op.drop_table('hosts')
    op.drop_index(op.f('ix_groups_zone_id'), table_name='groups')
    op.drop_index(op.f('ix_groups_parent_id'), table_name='groups')
    op.drop_index(op.f('ix_groups_name'), table_name='groups')
    op.drop_index('groups_name_zone_id_key_null', table_name='groups')
    op.drop_index('groups_name_zone_id_key_not_null', table_name='groups')
    op.drop_table('groups')
    op.drop_index(op.f('ix_zones_name'), table_name='zones')
    op.drop_table('zones')
    op.drop_index(op.f('ix_users_session_token'), table_name='users')
    op.drop_index(op.f('ix_users_name'), table_name='users')
    op.drop_index(op.f('ix_users_email'), table_name='users')
    op.drop_table('users')
    op.drop_index(op.f('ix_roles_name'), table_name='roles')
    op.drop_index(op.f('ix_roles_ldap_group'), table_name='roles')
    op.drop_table('roles')
    # ### end Alembic commands ###



from collections import OrderedDict
from opsy.inventory.schema import (
    ZoneSchema, ZoneQuerySchema, ZoneRefSchema,
    HostSchema, HostQuerySchema, HostRefSchema,
    GroupSchema, GroupQuerySchema, GroupRefSchema,
    HostGroupMappingSchema, HostGroupMappingQuerySchema,
    HostGroupMappingRefSchema)
from opsy.inventory.models import HostGroupMapping


###############################################################################
# ZoneSchema Tests
###############################################################################


def test_zone_schema(test_zone, test_zones):

    expected_zone_schema_output = OrderedDict([
        ('id', test_zone.id),
        ('name', test_zone.name),
        ('description', test_zone.description),
        ('vars', test_zone.vars),
        ('created_at', test_zone.created_at.isoformat()),
        ('updated_at', test_zone.updated_at.isoformat()),
        ('_links', {
            'self': f'/api/v1/zones/{test_zone.id}',
            'collection': '/api/v1/zones/'})])

    expected_zone_schema_input = OrderedDict([
        ('name', 'mytestzone')])

    expected_zone_ref_schema_output = OrderedDict([
        ('id', test_zone.id),
        ('name', test_zone.name),
        ('_links', {
            'self': f'/api/v1/zones/{test_zone.id}',
            'collection': '/api/v1/zones/'})])

    expected_zone_query_schema_output = OrderedDict([
        ('name', 'mytestzone')])

    assert ZoneSchema().dump(test_zone) == expected_zone_schema_output
    assert ZoneSchema().load(
        {'name': 'mytestzone'}) == expected_zone_schema_input
    assert ZoneRefSchema().dump(test_zone) == expected_zone_ref_schema_output
    assert ZoneQuerySchema(partial=True).dump(
        {'name': 'mytestzone'}) == expected_zone_query_schema_output


###############################################################################
# HostSchema Tests
###############################################################################


def test_host_schema(test_host, test_hosts):

    expected_host_schema_output = OrderedDict([
        ('id', test_host.id),
        ('zone_id', test_host.zone_id),
        ('name', test_host.name),
        ('vars', None),
        ('compiled_vars', {}),
        ('zone', OrderedDict([
            ('id', test_host.zone.id),
            ('name', test_host.zone.name),
            ('_links', {'self': f'/api/v1/zones/{test_host.zone_id}',
                        'collection': '/api/v1/zones/'})])),
        ('group_mappings', []),
        ('created_at', test_host.created_at.isoformat()),
        ('updated_at', test_host.updated_at.isoformat()),
        ('_links',
            {'self': f'/api/v1/hosts/{test_host.id}',
             'collection': '/api/v1/hosts/'})])

    expected_host_schema_input = OrderedDict([
        ('zone_id', test_host.zone_id),
        ('name', 'mytesthost')])

    expected_host_ref_schema_output = OrderedDict([
        ('id', test_host.id),
        ('name', test_host.name),
        ('_links', {
            'self': f'/api/v1/hosts/{test_host.id}',
            'collection': '/api/v1/hosts/'})])

    expected_host_query_schema_output = OrderedDict([
        ('name', 'mytesthost')])

    assert HostSchema().dump(test_host) == expected_host_schema_output
    assert HostSchema().load(
        {'name': 'mytesthost',
         'zone_id': test_host.zone_id}) == expected_host_schema_input
    assert HostRefSchema().dump(test_host) == expected_host_ref_schema_output
    assert HostQuerySchema(partial=True).dump(
        {'name': 'mytesthost'}) == expected_host_query_schema_output


###############################################################################
# GroupSchema Tests
###############################################################################


def test_group_schema(test_group, test_groups):

    expected_group_schema_output = OrderedDict([
        ('id', test_group.id),
        ('zone_id', test_group.zone_id),
        ('parent_id', test_group.parent_id),
        ('name', test_group.name),
        ('default_priority', test_group.default_priority),
        ('vars', test_group.vars),
        ('compiled_vars', test_group.compiled_vars),
        ('created_at', test_group.created_at.isoformat()),
        ('updated_at', test_group.updated_at.isoformat()),
        ('_links',
            {'self': f'/api/v1/groups/{test_group.id}',
             'collection': '/api/v1/groups/'})])

    expected_group_schema_input = OrderedDict([
        ('name', 'mytestgroup')])

    expected_group_ref_schema_output = OrderedDict([
        ('id', test_group.id),
        ('name', test_group.name),
        ('_links', {
            'self': f'/api/v1/groups/{test_group.id}',
            'collection': '/api/v1/groups/'})])

    expected_group_query_schema_output = OrderedDict([
        ('name', 'mytestgroup')])

    assert GroupSchema().dump(test_group) == expected_group_schema_output
    assert GroupSchema().load(
        {'name': 'mytestgroup'}) == expected_group_schema_input
    assert GroupRefSchema().dump(test_group) == expected_group_ref_schema_output
    assert GroupQuerySchema(partial=True).dump(
        {'name': 'mytestgroup'}) == expected_group_query_schema_output


###############################################################################
# HostGroupMappingSchema Tests
###############################################################################


def test_host_group_mapping_schema(test_inventory_bootstrap):

    test_mapping = HostGroupMapping.get_by_host_and_group(
        'westprom', 'prom_nodes')

    expected_host_group_mapping_schema_output = OrderedDict([
        ('id', test_mapping.id),
        ('host_id', test_mapping.host_id),
        ('group_id', test_mapping.group_id),
        ('priority', test_mapping.priority),
        ('host', OrderedDict([
            ('id', test_mapping.host.id),
            ('name', test_mapping.host.name),
            ('_links',
                {'self': f'/api/v1/hosts/{test_mapping.host.id}',
                 'collection': '/api/v1/hosts/'})])),
        ('group', OrderedDict([
            ('id', test_mapping.group.id),
            ('name', test_mapping.group.name),
            ('_links',
                {'self': f'/api/v1/groups/{test_mapping.group.id}',
                 'collection': '/api/v1/groups/'})])),
        ('created_at', test_mapping.created_at.isoformat()),
        ('updated_at', test_mapping.updated_at.isoformat()),
        ('_links',
            {'self': f'/api/v1/hosts/{test_mapping.host.id}'
                f'/group_mappings/{test_mapping.group.id}',
             'collection': f'/api/v1/hosts/{test_mapping.host.id}'
                '/group_mappings/'})])

    expected_host_group_mapping_ref_schema_output = OrderedDict([
        ('id', test_mapping.id),
        ('host_id', test_mapping.host_id),
        ('host_name', test_mapping.host_name),
        ('group_id', test_mapping.group_id),
        ('group_name', test_mapping.group_name),
        ('priority', test_mapping.priority),
        ('_links',
            {'self': f'/api/v1/hosts/{test_mapping.host.id}'
                f'/group_mappings/{test_mapping.group.id}',
             'collection': f'/api/v1/hosts/{test_mapping.host.id}'
                '/group_mappings/'})])

    expected_host_group_mapping_query_schema_output = OrderedDict([
        ('group_name', 'mytestmapping')])

    assert HostGroupMappingSchema().dump(test_mapping) == \
        expected_host_group_mapping_schema_output
    assert HostGroupMappingRefSchema().dump(test_mapping) == \
        expected_host_group_mapping_ref_schema_output
    assert HostGroupMappingQuerySchema().dump(
        {'group___name': 'mytestmapping'}) == \
        expected_host_group_mapping_query_schema_output



import os
import sys
from functools import partial
import click
from flask import current_app
from flask.cli import (
    AppGroup, routes_command, ScriptInfo, with_appcontext, pass_script_info)
from flask_migrate.cli import db as db_command

import opsy
from opsy.flask_extensions import db
from opsy.app import create_app
from opsy.config import load_config
from opsy.exceptions import NoConfigFile
from opsy.server import create_server
from opsy.auth.schema import AppPermissionSchema, UserSchema, RoleSchema
from opsy.utils import (
    print_error, print_notice, get_protected_routes, get_valid_permissions)
from opsy.auth.models import Role, User, Permission
from opsy.inventory.models import Zone, Host, Group, HostGroupMapping


click_option = partial(  # pylint: disable=invalid-name
    click.option, show_default=True, show_envvar=True)


@click.group(cls=AppGroup, help='The Opsy management cli.')
@click_option('--config', type=click.Path(),
              default=f'{os.path.abspath(os.path.curdir)}/opsy.toml',
              envvar='OPSY_CONFIG', help='Config file for opsy.')
@click.pass_context
def cli(ctx, config):
    try:
        ctx.obj.data['config'] = load_config(config)
    except NoConfigFile as error:
        print_error(error, exit_script=False)
        ctx.obj.data['config'] = None


cli.add_command(routes_command)
cli.add_command(db_command)


@cli.command('run')
@click_option('--host', type=click.STRING, help='Host address.')
@click_option('--port', '-p', type=click.INT, help='Port number to listen on.')
@click_option('--threads', '-t', type=click.INT, help='Amount of threads.')
@click_option('--ssl_enabled', type=click.BOOL, is_flag=True)
@click_option('--certificate', type=click.Path(), help='SSL cert.')
@click_option('--private_key', type=click.Path(), help='SSL key.')
@click_option('--ca_certificate', type=click.Path(), help='SSL CA cert.')
@pass_script_info
def run(script_info, host, port, threads, ssl_enabled, certificate,
        private_key, ca_certificate):
    """Run the Opsy server."""
    app = script_info.load_app()
    host = host or app.config.opsy['server']['host']
    port = port or app.config.opsy['server']['port']
    threads = threads or app.config.opsy['server']['threads']
    ssl_enabled = ssl_enabled or app.config.opsy['server']['ssl_enabled']
    certificate = certificate or app.config.opsy['server']['certificate']
    private_key = private_key or app.config.opsy['server']['private_key']
    ca_certificate = ca_certificate or \
        app.config.opsy['server']['ca_certificate']
    server = create_server(app, host, port, threads, ssl_enabled, certificate,
                           private_key, ca_certificate)
    try:
        proto = 'https' if server.ssl_adapter else 'http'
        app.logger.info(f'Starting Opsy server at {proto}://{host}:{port}/...')
        app.logger.info(f'API docs available at {proto}://{host}:{port}/docs/')
        server.start()
    except KeyboardInterrupt:
        app.logger.info('Stopping Opsy server...')
    finally:
        server.stop()


@cli.command('shell')
def shell():
    """Run a shell in the app context."""
    from flask.globals import _app_ctx_stack
    banner = 'Welcome to Opsy!'
    app = _app_ctx_stack.top.app
    shell_ctx = {'create_app': create_app,
                 'db': db,
                 'User': User,
                 'Role': Role,
                 'Permission': Permission,
                 'Zone': Zone,
                 'Host': Host,
                 'Group': Group,
                 'HostGroupMapping': HostGroupMapping}
    shell_ctx.update(app.make_shell_context())
    try:
        from IPython import embed
        embed(user_ns=shell_ctx, banner1=banner)
        return
    except ImportError:
        import code
        code.interact(banner, local=shell_ctx)


@db_command.command('init-db')
@click.confirmation_option(
    prompt='This will delete everything. Do you want to continue?')
@with_appcontext
def init_db():
    """Drop everything in database and rebuild the schema."""
    current_app.logger.info('Creating database...')
    db.drop_all()
    db.create_all()
    db.session.commit()


@cli.command('permission-list')
@click_option('--resource', type=click.STRING)
@click_option('--method', type=click.STRING)
def permission_list(**kwargs):
    """List all permissions the app is aware of."""
    print(AppPermissionSchema(many=True).dumps(
        get_protected_routes(ignored_methods=["HEAD", "OPTIONS"]), indent=4))


@cli.command('create-admin-user')
@click_option('--password', '-p', hide_input=True, confirmation_prompt=True,
              envvar='OPSY_ADMIN_PASSWORD',
              prompt='Password for the new admin user',
              help='Password for the new admin user.')
@click_option('--force', '-f', type=click.BOOL, is_flag=True,
              help='Recreate admin user and role if they already exist.')
def create_admin_user(password, force):
    """Create the default admin user."""
    admin_user = User.query.filter_by(name='admin').first()
    admin_role = Role.query.filter_by(name='admin').first()
    if admin_user and not force:
        print_notice('Admin user already found, exiting. '
                     'Use "--force" to force recreation.')
        sys.exit(0)
    if admin_role and not force:
        print_notice('Admin role already found, exiting. '
                     'Use "--force" to force recreation.')
        sys.exit(0)
    if admin_user:
        print_notice('Admin user already found, deleting.')
        admin_user.delete()
    if admin_role:
        print_notice('Admin role already found, deleting.')
        admin_role.delete()
    admin_user = User.create(
        'admin', password=password, full_name='Default admin user')
    admin_role = Role.create('admin', description='Default admin role')
    for permission in get_valid_permissions():
        admin_role.add_permission(permission)
    admin_role.add_user(admin_user)
    admin_role.save()
    admin_user = User.query.filter_by(name='admin').first()
    admin_role = Role.query.filter_by(name='admin').first()
    print_notice('Admin user created with the specified password:')
    print(UserSchema().dumps(admin_user, indent=4))
    print_notice('Admin role created:')
    print(RoleSchema().dumps(admin_role, indent=4))


@cli.command('version', with_appcontext=False)
def version():
    """Just show the version and quit."""
    print(opsy.__version__)


def main():

    def create_opsy_app(script_info):
        if not script_info.data['config']:
            print_error('Config file not loaded, unable to start app.')
        return create_app(script_info.data['config'])
    cli(  # pylint: disable=unexpected-keyword-arg,no-value-for-parameter
        obj=ScriptInfo(create_app=create_opsy_app))



from flask import Flask
from opsy.config import configure_app
from opsy.flask_extensions import configure_extensions, finalize_extensions
from opsy.logging import (configure_logging, log_before_request,
                          log_after_request)
from opsy.auth.views import create_auth_views
from opsy.inventory.views import create_inventory_views


class PrefixMiddleware:

    def __init__(self, app, prefix=''):
        self.app = app
        self.prefix = prefix

    def __call__(self, environ, start_response):
        if environ['PATH_INFO'].startswith(self.prefix):
            environ['PATH_INFO'] = environ['PATH_INFO'][len(self.prefix):]
            environ['SCRIPT_NAME'] = self.prefix
            return self.app(environ, start_response)
        start_response('404', [('Content-Type', 'text/plain')])
        return ["This url does not belong to the app.".encode()]


def create_app(config):
    configure_logging(config)
    app = Flask('opsy')
    app.wsgi_app = PrefixMiddleware(
        app.wsgi_app, prefix=config['app']['uri_prefix'])
    app.before_request(log_before_request)
    app.after_request(log_after_request)
    configure_app(app, config)
    configure_extensions(app)
    create_views(app)
    finalize_extensions(app)
    return app


def create_views(app):
    create_auth_views(app)
    create_inventory_views(app)



import json
from opsy.auth.utils import create_token
from opsy.inventory.models import Zone, Host, Group, HostGroupMapping

###############################################################################
# Zone Tests
###############################################################################


def test_zones_list(client, admin_user, test_user, test_inventory_bootstrap):
    # First we just create the token for both of our users.
    create_token(admin_user)
    create_token(test_user)
    # Then we use this token to list the zones for both users.
    admin_user_response = client.get(
        '/api/v1/zones/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    user_response = client.get(
        '/api/v1/zones/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    # Then we make sure we get the proper response codes.
    assert admin_user_response.status_code == 200
    assert user_response.status_code == 403
    # Now we parse the admin data.
    admin_response_data = json.loads(admin_user_response.data)
    # We should have 3 zones.
    assert len(admin_response_data) == 3
    # Let's test again with a filter.
    admin_user_response = client.get(
        '/api/v1/zones/?name=west',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    admin_response_data = json.loads(admin_user_response.data)
    # Now we should only have 1 zone named west.
    assert len(admin_response_data) == 1
    assert admin_response_data[0]['name'] == 'west'


def test_zones_post(client, admin_user, test_user, test_inventory_bootstrap):
    # Zone test data
    zone_data = {
        "name": "test zone",
        "description": "this is only a test",
        "vars": {}
    }
    # Create tokens
    create_token(test_user)
    create_token(admin_user)
    # Test unprivileged user zone creation fails
    user_zone_create = client.post(
        '/api/v1/zones/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)],
        json=zone_data)
    assert user_zone_create.status_code == 403
    # Test admin user zone creation succeeds
    admin_zone_create = client.post(
        '/api/v1/zones/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json=zone_data)
    assert admin_zone_create.status_code == 201
    # Verify admin user zone creation was successful
    admin_zone_verify = client.get(
        f"/api/v1/zones/{zone_data['name']}",
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    admin_zone_verify_data = json.loads(admin_zone_verify.data)
    assert admin_zone_verify_data['name'] == zone_data['name']
    # Verify duplicate zone insertion fails
    duplicate_insert = client.post(
        '/api/v1/zones/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json=zone_data)
    assert duplicate_insert.status_code == 400


def test_zones_get(client, admin_user, test_user, test_inventory_bootstrap):
    # Create tokens
    create_token(test_user)
    create_token(admin_user)
    # Test unprivileged user zone get by name fails
    user_zone_get_by_name = client.get(
        '/api/v1/zones/west',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert user_zone_get_by_name.status_code == 403
    # Test admin user zone get by name succeeds
    admin_zone_get_by_name = client.get(
        '/api/v1/zones/west',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert admin_zone_get_by_name.status_code == 200
    admin_zone_get_by_name_data = json.loads(admin_zone_get_by_name.data)
    assert admin_zone_get_by_name_data['name'] == 'west'
    # Get a zone id to test
    zone_id = admin_zone_get_by_name_data['id']
    # Test unprivileged user get zone by id fails
    user_zone_get_by_id = client.get(
        f'/api/v1/zones/{zone_id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert user_zone_get_by_id.status_code == 403
    # Test admin user zone get by id succeeds
    admin_zone_get_by_id = client.get(
        f'/api/v1/zones/{zone_id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert admin_zone_get_by_id.status_code == 200
    admin_zone_get_by_id_data = json.loads(admin_zone_get_by_id.data)
    assert admin_zone_get_by_id_data['name'] == 'west'
    # Test getting by name or id returns same data
    test_zone = Zone.get_by_id_or_name('west')
    zone_get_by_name = client.get(
        f'/api/v1/zones/{test_zone.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    zone_get_by_id = client.get(
        f'/api/v1/zones/{test_zone.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert zone_get_by_name.data == zone_get_by_id.data


def test_zones_patch(client, admin_user, test_user, test_inventory_bootstrap):
    # Create tokens
    create_token(test_user)
    create_token(admin_user)
    # Test that patching nonexistent zone fails
    missing_zone_patch_by_name = client.patch(
        '/api/v1/zones/this_zone_should_not_exist',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json={"description": "nonexistent zone"})
    assert missing_zone_patch_by_name.status_code == 404
    missing_zone_patch_by_id = client.patch(
        'api/v1/zones/12345',  # that's the combination on my luggage!
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json={"description": "nonexistent zone"})
    assert missing_zone_patch_by_id.status_code == 404
    # Test that patching existing zone by admin succeeds
    zone_update_by_name = {"description": "updated by name"}
    zone_update_by_id = {"description": "updated by id"}
    # Patching by name
    admin_zone_patch_by_name = client.patch(
        '/api/v1/zones/west',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json=zone_update_by_name)
    assert admin_zone_patch_by_name.status_code == 200
    get_updated_zone_by_name = client.get(
        '/api/v1/zones/west',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    get_updated_zone_by_name_data = json.loads(get_updated_zone_by_name.data)
    assert get_updated_zone_by_name_data['description'] == zone_update_by_name['description']
    # Patching by id
    zone_id = get_updated_zone_by_name_data['id']
    admin_zone_patch_by_id = client.patch(
        f'/api/v1/zones/{zone_id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json=zone_update_by_id)
    assert admin_zone_patch_by_id.status_code == 200
    get_updated_zone_by_id = client.get(
        f'/api/v1/zones/{zone_id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    get_updated_zone_by_id_data = json.loads(get_updated_zone_by_id.data)
    assert get_updated_zone_by_id_data['description'] == zone_update_by_id['description']
    # Test that patching an existing zone by unprivileged user fails
    user_zone_patch_by_name = client.patch(
        '/api/v1/zones/west',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)],
        json={'description': 'updated'})
    assert user_zone_patch_by_name.status_code == 403
    user_zone_patch_by_id = client.patch(
        f'/api/v1/zones/{zone_id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)],
        json={'description': 'updated'})
    assert user_zone_patch_by_id.status_code == 403


def test_zones_delete(client, admin_user, test_user, test_inventory_bootstrap):
    # Create tokens
    create_token(test_user)
    create_token(admin_user)
    # Test that deleting nonexistent zone fails
    delete_missing_name = client.delete(
        '/api/v1/zones/southbysouthwest',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert delete_missing_name.status_code == 404
    delete_missing_id = client.delete(
        '/api/v1/zones/12345',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert delete_missing_id.status_code == 404
    # Test that unprivileged user cannot delete zones
    delete_nopriv_name = client.delete(
        '/api/v1/zones/southbysouthwest',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert delete_nopriv_name.status_code == 403
    delete_nopriv_id = client.delete(
        '/api/v1/zones/12345',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert delete_nopriv_id.status_code == 403
    # Test that admin can delete zones
    delete_admin_name = client.delete(
        '/api/v1/zones/west',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert delete_admin_name.status_code == 204
    delete_admin_name_verify = client.get(
        '/api/v1/zones/west',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert delete_admin_name_verify.status_code == 404
    test_zone = Zone.get_by_id_or_name('east')
    delete_admin_id = client.delete(
        f'/api/v1/zones/{test_zone.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert delete_admin_id.status_code == 204
    delete_admin_id_verify = client.get(
        f'/api/v1/zones/{test_zone.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert delete_admin_id_verify == 404


###############################################################################
# Host Tests
###############################################################################

def test_hosts_list(client, admin_user, test_user, test_inventory_bootstrap):
    # Set up tokens
    create_token(admin_user)
    create_token(test_user)
    # Test that unprivileged user cannot list hosts
    list_hosts_nopriv = client.get(
        '/api/v1/hosts',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert list_hosts_nopriv.status_code == 403
    # Test that admin can list hosts
    list_hosts_admin = client.get(
        '/api/v1/hosts',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert list_hosts_admin.status_code == 200
    hosts_payload = json.loads(list_hosts_admin.data)
    assert len(hosts_payload) == 6
    # Test again with filter
    list_hosts_filter = client.get(
        '/api/v1/hosts/?name=westconsul',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert list_hosts_filter.status_code == 200
    list_hosts_filter_out = json.loads(list_hosts_filter.data)
    assert len(list_hosts_filter_out) == 1
    assert list_hosts_filter_out[0]['name'] == 'westconsul'


def test_hosts_post(client, admin_user, test_user, test_inventory_bootstrap):
    # Host test data
    test_host = {
        "zone_id": Zone.get_by_id_or_name('west').id,
        "name": "test host"
    }
    # Create tokens
    create_token(admin_user)
    create_token(test_user)
    # Test that a nopriv user cannot create a host
    host_create_nopriv = client.post(
        '/api/v1/hosts/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)],
        json=test_host)
    assert host_create_nopriv.status_code == 403
    # Test that admin can create a host
    host_create_admin = client.post(
        '/api/v1/hosts/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json=test_host)
    assert host_create_admin.status_code == 201
    # Verify host was created correctly
    host_verify_admin = client.get(
        f"/api/v1/hosts/{test_host['name']}",
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    verify_out = json.loads(host_verify_admin.data)
    assert verify_out['name'] == test_host['name']
    # Test that creating a duplicate host fails
    host_dup_insert = client.post(
        '/api/v1/hosts/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json=test_host)
    assert host_dup_insert.status_code == 400


def test_hosts_get(client, admin_user, test_user, test_inventory_bootstrap):
    # Create tokens
    create_token(test_user)
    create_token(admin_user)
    # Get a host for testing
    test_host = Host.get_by_id_or_name('westconsul')
    # Test that a nopriv user cannot get hosts
    host_nopriv_get_name = client.get(
        f'/api/v1/hosts/{test_host.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert host_nopriv_get_name.status_code == 403
    host_nopriv_get_id = client.get(
        f'/api/v1/hosts/{test_host.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert host_nopriv_get_id.status_code == 403
    # Test that admin can get hosts
    host_admin_get_name = client.get(
        f'/api/v1/hosts/{test_host.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert host_admin_get_name.status_code == 200
    name_out = json.loads(host_admin_get_name.data)
    assert name_out['id'] == test_host.id
    host_admin_get_id = client.get(
        f'/api/v1/hosts/{test_host.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert host_admin_get_id.status_code == 200
    id_out = json.loads(host_admin_get_id.data)
    assert id_out['name'] == test_host.name
    # Test that hosts_get by name or id fails on nonexistent host
    fail_get_id = client.get(
        '/api/v1/hosts/12345',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert fail_get_id.status_code == 404
    fail_get_name = client.get(
        '/api/v1/hosts/asdfhjkl',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert fail_get_name.status_code == 404


def test_hosts_patch(client, admin_user, test_user, test_inventory_bootstrap):
    # Test data
    test_data = {"vars": {"deadbeef": True}}
    test_host = Host.get_by_id_or_name('westconsul')
    # Create tokens
    create_token(test_user)
    create_token(admin_user)
    # Test that patching a nonexistent host fails
    nohost_patch_name = client.patch(
        '/api/v1/hosts/this_host_should_not_exist',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json=test_data)
    assert nohost_patch_name.status_code == 404
    nohost_patch_id = client.patch(
        '/api/v1/hosts/12345',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json=test_data)
    assert nohost_patch_id.status_code == 404
    # Test that patching by nopriv user fails
    nopriv_host_name = client.patch(
        f'/api/v1/hosts/{test_host.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)],
        json=test_data)
    assert nopriv_host_name.status_code == 403
    nopriv_host_id = client.patch(
        f'/api/v1/hosts/{test_host.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)],
        json=test_data)
    assert nopriv_host_id.status_code == 403
    # Test that patching by admin succeeds
    admin_host_name = client.patch(
        f'/api/v1/hosts/{test_host.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json=test_data)
    assert admin_host_name.status_code == 200
    admin_verify_name = client.get(
        f'/api/v1/hosts/{test_host.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    name_out = json.loads(admin_verify_name.data)
    assert name_out['vars'] == test_data['vars']
    admin_host_id = client.patch(
        f'/api/v1/hosts/{test_host.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json={"vars": {}})
    assert admin_host_id.status_code == 200
    admin_verify_id = client.get(
        f'/api/v1/hosts/{test_host.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    id_out = json.loads(admin_verify_id.data)
    assert id_out['vars'] == {}


def test_hosts_delete(client, admin_user, test_user, test_inventory_bootstrap):
    # Test data
    test_hosts = [Host.get_by_id_or_name('westconsul'),
                  Host.get_by_id_or_name('eastconsul')]
    # Create tokens
    create_token(test_user)
    create_token(admin_user)
    # Test that deleting a nonexistent host fails
    nohost_delete_name = client.delete(
        '/api/v1/hosts/not_a_host',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert nohost_delete_name.status_code == 404
    nohost_delete_id = client.delete(
        '/api/v1/hosts/12345',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert nohost_delete_id.status_code == 404
    # Test that nopriv user cannot delete hosts
    nopriv_delete_name = client.delete(
        f'/api/v1/hosts/{test_hosts[0].name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert nopriv_delete_name.status_code == 403
    nopriv_delete_id = client.delete(
        f'/api/v1/hosts/{test_hosts[0].id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert nopriv_delete_id.status_code == 403
    # Test that admin can delete hosts by name
    admin_delete_name = client.delete(
        f'/api/v1/hosts/{test_hosts[0].name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert admin_delete_name.status_code == 204
    verify_name = client.get(
        f'/api/v1/hosts/{test_hosts[0].name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert verify_name.status_code == 404
    # Test that admin can delete hosts by id
    admin_delete_id = client.delete(
        f'/api/v1/hosts/{test_hosts[1].id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert admin_delete_id.status_code == 204
    verify_id = client.get(
        f'/api/v1/hosts/{test_hosts[1].id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert verify_id.status_code == 404


###############################################################################
# Host Group Mapping Tests
###############################################################################


def test_host_group_mappings_list(client, test_user, admin_user, test_inventory_bootstrap):
    # Test data
    test_host = Host.get_by_id_or_name('westprom')
    # Create tokens
    create_token(test_user)
    create_token(admin_user)
    # Test that listing nonexistent host fails
    nohost_list_name = client.get(
        '/api/v1/hosts/not_a_host/group_mappings',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert nohost_list_name.status_code == 404
    nohost_list_id = client.get(
        '/api/v1/hosts/12345/group_mappings',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert nohost_list_id.status_code == 404
    # Test that listing by nopriv user fails
    nopriv_list_name = client.get(
        f'/api/v1/hosts/{test_host.name}/group_mappings',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert nopriv_list_name.status_code == 403
    nopriv_list_id = client.get(
        f'/api/v1/hosts/{test_host.id}/group_mappings',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert nopriv_list_id.status_code == 403
    # Test that listing by admin succeeds
    admin_list_name = client.get(
        f'/api/v1/hosts/{test_host.name}/group_mappings',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert admin_list_name.status_code == 200
    name_out = json.loads(admin_list_name.data)
    api_ids = sorted([x['id'] for x in name_out])
    host_ids = sorted([y.id for y in test_host.group_mappings])
    assert host_ids == api_ids
    # and by id
    admin_list_id = client.get(
        f'/api/v1/hosts/{test_host.id}/group_mappings',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert admin_list_id.status_code == 200
    id_out = sorted([z['id'] for z in json.loads(admin_list_id.data)])
    assert id_out == host_ids
    # Test again with a filter
    admin_filter = client.get(
        f'/api/v1/hosts/{test_host.name}/group_mappings/?id={test_host.group_mappings[0].id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert admin_filter.status_code == 200
    admin_filter_out = json.loads(admin_filter.data)
    assert len(admin_filter_out) == 1
    assert admin_filter_out[0]['id'] == test_host.group_mappings[0].id


def test_host_group_mappings_post(client, test_user, admin_user, test_inventory_bootstrap):
    # Test data
    zone = Zone.get_by_id_or_name('west')
    test_group = Group.create(name='test group', zone=zone)
    test_host = Host.get_by_id_or_name('westprom')
    another_host = Host.get_by_id_or_name('westconsul')
    # Create tokens
    create_token(test_user)
    create_token(admin_user)
    # Test that adding a group mapping to a nonexistent host fails
    nohost_mapping_name = client.post(
        '/api/v1/hosts/not_a_host/group_mappings',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json={"group_id": test_group.id})
    assert nohost_mapping_name.status_code == 404
    # by id as well
    nohost_mapping_id = client.post(
        '/api/v1/hosts/12345/group_mappings',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json={"group_id": test_group.id})
    assert nohost_mapping_id.status_code == 404
    # Test that mapping a nonexistent group to a host fails
    nogroup_mapping_name = client.post(
        f'/api/v1/hosts/{test_host.name}/group_mappings',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json={"group_id": "12345"})
    assert nogroup_mapping_name.status_code == 404
    # also by id
    nogroup_mapping_id = client.post(
        f'/api/v1/hosts/{test_host.id}/group_mappings',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json={"group_id": "12345"})
    assert nogroup_mapping_id.status_code == 404
    # Test that adding a group to a host by nopriv user fails
    nopriv_mapping_name = client.post(
        f'/api/v1/hosts/{test_host.name}/group_mappings',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)],
        json={"group_id": test_group.id})
    assert nopriv_mapping_name.status_code == 403
    # nopriv by id
    nopriv_mapping_id = client.post(
        f'/api/v1/hosts/{test_host.id}/group_mappings',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)],
        json={"group_id": test_group.id})
    assert nopriv_mapping_id.status_code == 403
    # Try it with an invalid group as well
    nopriv_mapping_nogroup = client.post(
        f'/api/v1/hosts/{test_host.name}/group_mappings',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)],
        json={"group_id": "12345"})
    assert nopriv_mapping_nogroup.status_code == 403  # not 404
    # Test that admin can add a group mapping
    admin_mapping_name = client.post(
        f'/api/v1/hosts/{test_host.name}/group_mappings',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json={"group_id": f"{test_group.id}"})
    assert admin_mapping_name.status_code == 201
    verify_name = client.get(
        f'/api/v1/hosts/{test_host.name}/group_mappings',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    verify_name_list = [x['id'] for x in json.loads(verify_name.data)]
    mapping1 = HostGroupMapping.query.filter_by(
        group_id=test_group.id,
        host_id=test_host.id).first()
    assert mapping1.id in verify_name_list
    # Again, with id
    admin_mapping_id = client.post(
        f'/api/v1/hosts/{another_host.id}/group_mappings',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json={"group_id": test_group.id})
    assert admin_mapping_id.status_code == 201
    verify_id = client.get(
        f'/api/v1/hosts/{another_host.id}/group_mappings',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    verify_id_list = [y['id'] for y in json.loads(verify_id.data)]
    mapping2 = HostGroupMapping.query.filter_by(
        group_id=test_group.id,
        host_id=another_host.id).first()
    assert mapping2.id in verify_id_list
    # Test that adding a duplicate fails
    dup_mapping = client.post(
        f'/api/v1/hosts/{another_host.name}/group_mappings',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json={"group_id": test_group.id})
    assert dup_mapping.status_code == 400


def test_host_group_mappings_get(client, test_user, admin_user, test_inventory_bootstrap):
    # Test data
    test_host = Host.get_by_id_or_name('westprom')
    test_group = test_host.groups[0]
    test_mapping = HostGroupMapping.get_by_host_and_group(test_host.id, test_group.id)
    # Create tokens
    create_token(test_user)
    create_token(admin_user)
    # Test that invalid host name fails with valid mapping
    nohost_get_name = client.get(
        f'/api/v1/hosts/not_a_host/group_mappings/{test_group.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert nohost_get_name.status_code == 404
    # Test that invalid host id also fails with valid mapping
    nohost_get_id = client.get(
        f'/api/v1/hosts/12345/group_mappings/{test_group.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert nohost_get_id.status_code == 404
    # by group name as well
    nohost_get_id_groupname = client.get(
        f'/api/v1/hosts/12345/group_mappings/{test_group.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert nohost_get_id_groupname.status_code == 404
    # Test that invalid host name with valid group name fails
    nohost_get_name_groupname = client.get(
        f'/api/v1/hosts/not_a_host/group_mappings/{test_group.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert nohost_get_name_groupname.status_code == 404
    # Test invalid host with invalid mapping
    nohost_get_nomap = client.get(
        f'/api/v1/hosts/not_a_host/group_mappings/12345',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert nohost_get_nomap.status_code == 404
    # By id as well
    nohost_get_nomap_id = client.get(
        f'/api/v1/hosts/98765/group_mappings/asdf',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert nohost_get_nomap_id.status_code == 404
    # Test that nopriv user cannot get mappings
    nopriv_get_name = client.get(
        f'/api/v1/hosts/{test_host.name}/group_mappings/{test_group.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert nopriv_get_name.status_code == 403
    # Test by id as well
    nopriv_get_id = client.get(
        f'/api/v1/hosts/{test_host.id}/group_mappings/{test_group.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert nopriv_get_id.status_code == 403
    # Test by group name as well
    gn_nopriv_get_name = client.get(
        f'/api/v1/hosts/{test_host.name}/group_mappings/{test_group.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert gn_nopriv_get_name.status_code == 403
    # Test nopriv by host id with group name
    gn_nopriv_get_id = client.get(
        f'/api/v1/hosts/{test_host.id}/group_mappings/{test_group.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert gn_nopriv_get_id.status_code == 403
    # Test that invalid host with nopriv user fails correctly
    nopriv_nohost = client.get(
        f'/api/v1/hosts/not_a_host/group_mappings/{test_group.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert nopriv_nohost.status_code == 403  # not 404
    # Test that invalid mapping with nopriv user fails correctly
    nopriv_nomap = client.get(
        f'/api/v1/hosts/{test_host.id}/group_mappings/12345',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert nopriv_nomap.status_code == 403  # not 404
    # Test that admin can get mappings by host name
    admin_get_name = client.get(
        f'/api/v1/hosts/{test_host.name}/group_mappings/{test_group.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert admin_get_name.status_code == 200
    name_out = json.loads(admin_get_name.data)
    assert name_out['id'] == test_mapping.id
    # and by id
    admin_get_id = client.get(
        f'/api/v1/hosts/{test_host.id}/group_mappings/{test_group.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert admin_get_id.status_code == 200
    id_out = json.loads(admin_get_id.data)
    assert id_out['id'] == test_mapping.id
    # and by group name
    admin_get_group_name = client.get(
        f'/api/v1/hosts/{test_host.name}/group_mappings/{test_group.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert admin_get_group_name.status_code == 200
    group_name_out = json.loads(admin_get_group_name.data)
    assert group_name_out['id'] == test_mapping.id
    # once more, with host id and group name
    admin_get_group_id = client.get(
        f'/api/v1/hosts/{test_host.id}/group_mappings/{test_group.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert admin_get_group_id.status_code == 200
    group_id_out = json.loads(admin_get_group_id.data)
    assert group_id_out['id'] == test_mapping.id


def test_host_group_mappings_patch(client, test_user, admin_user, test_inventory_bootstrap):
    # Test data
    test_host = Host.get_by_id_or_name('westprom')
    test_group = test_host.groups[0]
    test_data = {"priority": 9}
    alt_data = {"priority": 8}
    # Create tokens
    create_token(test_user)
    create_token(admin_user)
    # Test invalid data - invalid host id
    invalid_host_id = client.patch(
        f'/api/v1/hosts/12345/group_mappings/{test_group.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json=test_data)
    assert invalid_host_id.status_code == 404
    # Invalid host name
    invalid_host_name = client.patch(
        f'/api/v1/hosts/not_a_host/group_mappings/{test_group.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json=test_data)
    assert invalid_host_name.status_code == 404
    # Invalid group id
    invalid_group_id = client.patch(
        f'/api/v1/hosts/{test_host.name}/group_mappings/12345',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json=test_data)
    assert invalid_group_id.status_code == 404
    # Invalid group name
    invalid_group_name = client.patch(
        f'/api/v1/hosts/{test_host.id}/group_mappings/not_a_group',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json=test_data)
    assert invalid_group_name.status_code == 404
    # Test nopriv user - host name, group name
    nopriv_name_name = client.patch(
        f'/api/v1/hosts/{test_host.name}/group_mappings/{test_group.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)],
        json=test_data)
    assert nopriv_name_name.status_code == 403
    # nopriv - host name, group id
    nopriv_name_id = client.patch(
        f'/api/v1/hosts/{test_host.name}/group_mappings/{test_group.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)],
        json=test_data)
    assert nopriv_name_id.status_code == 403
    # nopriv - host id, group name
    nopriv_id_name = client.patch(
        f'/api/v1/hosts/{test_host.id}/group_mappings/{test_group.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)],
        json=test_data)
    assert nopriv_id_name.status_code == 403
    # nopriv - host id, group id
    nopriv_id_id = client.patch(
        f'/api/v1/hosts/{test_host.id}/group_mappings/{test_group.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)],
        json=test_data)
    assert nopriv_id_id.status_code == 403
    # Test admin user - host name, group name
    admin_name_name = client.patch(
        f'/api/v1/hosts/{test_host.name}/group_mappings/{test_group.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json=test_data)
    assert admin_name_name.status_code == 200
    output_name_name = json.loads(admin_name_name.data)
    assert output_name_name['priority'] == test_data['priority']
    # admin - host name, group id
    admin_name_id = client.patch(
        f'/api/v1/hosts/{test_host.name}/group_mappings/{test_group.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json=alt_data)
    assert admin_name_id.status_code == 200
    output_name_id = json.loads(admin_name_id.data)
    assert output_name_id['priority'] == alt_data['priority']
    # admin - host id, group name
    admin_id_name = client.patch(
        f'/api/v1/hosts/{test_host.id}/group_mappings/{test_group.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json=test_data)
    assert admin_id_name.status_code == 200
    output_id_name = json.loads(admin_id_name.data)
    assert output_id_name['priority'] == test_data['priority']
    # admin - host id, group id
    admin_id_id = client.patch(
        f'/api/v1/hosts/{test_host.id}/group_mappings/{test_group.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json=alt_data)
    assert admin_id_id.status_code == 200
    output_id_id = json.loads(admin_id_id.data)
    assert output_id_id['priority'] == alt_data['priority']


def test_host_group_mappings_delete(client, test_user, admin_user, test_inventory_bootstrap):
    # Test all the data
    test_host1 = Host.get_by_id_or_name('westprom')
    test_group1 = test_host1.groups[0]
    test_group2 = test_host1.groups[1]
    test_mapping1 = HostGroupMapping.get_by_host_and_group(test_host1.id, test_group1.id)
    test_mapping2 = HostGroupMapping.get_by_host_and_group(test_host1.id, test_group2.id)
    test_host2 = Host.get_by_id_or_name('westconsul')
    test_group3 = test_host2.groups[0]
    test_mapping3 = HostGroupMapping.get_by_host_and_group(test_host2.id, test_group3.id)
    test_host3 = Host.get_by_id_or_name('eastprom')
    test_group4 = test_host3.groups[0]
    test_mapping4 = HostGroupMapping.get_by_host_and_group(test_host3.id, test_group4.id)
    # Create tokens
    create_token(test_user)
    create_token(admin_user)
    # Test invalid data - host name, group name
    invalid_name_name = client.delete(
        '/api/v1/hosts/not_a_host/group_mappings/not_a_group',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert invalid_name_name.status_code == 404
    # Test invalid data - host name, group id
    invalid_name_id = client.delete(
        '/api/v1/hosts/not_a_host/group_mappings/12345',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert invalid_name_id.status_code == 404
    # Test invalid data - host id, group name
    invalid_id_name = client.delete(
        '/api/v1/hosts/12345/group_mappings/not_a_group',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert invalid_id_name.status_code == 404
    # Test invalid data - host id, group id
    invalid_id_id = client.delete(
        '/api/v1/hosts/12345/group_mappings/98765',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert invalid_id_id.status_code == 404
    # Test nopriv user - host name, group name
    nopriv_name_name = client.delete(
        f'/api/v1/hosts/{test_host1.name}/group_mappings/{test_group1.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert nopriv_name_name.status_code == 403
    # Test nopriv user - host name, group id
    nopriv_name_id = client.delete(
        f'/api/v1/hosts/{test_host1.name}/group_mappings/{test_group1.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert nopriv_name_id.status_code == 403
    # Test nopriv user - host id, group name
    nopriv_id_name = client.delete(
        f'/api/v1/hosts/{test_host1.id}/group_mappings/{test_group1.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert nopriv_id_name.status_code == 403
    # Test nopriv user - host id, group id
    nopriv_id_id = client.delete(
        f'/api/v1/hosts/{test_host1.id}/group_mappings/{test_group1.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert nopriv_id_id.status_code == 403
    # Test admin user - host name, group name
    admin_name_name = client.delete(
        f'/api/v1/hosts/{test_host1.name}/group_mappings/{test_group1.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert admin_name_name.status_code == 204
    assert HostGroupMapping.query.filter_by(id=test_mapping1.id).first() is None
    # Test admin user - host name, group id
    admin_name_id = client.delete(
        f'/api/v1/hosts/{test_host1.name}/group_mappings/{test_group2.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert admin_name_id.status_code == 204
    assert HostGroupMapping.query.filter_by(id=test_mapping2.id).first() is None
    # Test admin user - host id, group name
    admin_id_name = client.delete(
        f'/api/v1/hosts/{test_host2.id}/group_mappings/{test_group3.name}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert admin_id_name.status_code == 204
    assert HostGroupMapping.query.filter_by(id=test_mapping3.id).first() is None
    # Test admin user - host id, group id
    admin_id_id = client.delete(
        f'/api/v1/hosts/{test_host3.id}/group_mappings/{test_group4.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert admin_id_id.status_code == 204
    assert HostGroupMapping.query.filter_by(id=test_mapping4.id).first() is None


###############################################################################
# Group Tests
###############################################################################


def test_groups_list(client, test_user, admin_user, test_inventory_bootstrap):
    # Create tokens
    create_token(test_user)
    create_token(admin_user)
    # Test that nopriv user cannot list groups
    nopriv_list = client.get(
        '/api/v1/groups/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert nopriv_list.status_code == 403
    # Test that admin user can list groups
    admin_list = client.get(
        '/api/v1/groups/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert admin_list.status_code == 200
    # there should be 5 groups
    assert len(json.loads(admin_list.data)) == 5
    # Test with a filter as well
    admin_filter = client.get(
        '/api/v1/groups/?name=default',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert admin_filter.status_code == 200
    assert len(json.loads(admin_filter.data)) == 1


def test_groups_post(client, test_user, admin_user, test_inventory_bootstrap):
    # Create tokens
    create_token(test_user)
    create_token(admin_user)
    # Test that nopriv user cannot create a group
    nopriv_post = client.post(
        '/api/v1/groups/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)],
        json={"name": "test group"})
    assert nopriv_post.status_code == 403
    # Test that admin user can create a group
    admin_post = client.post(
        '/api/v1/groups/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json={"name": "test group", "vars": {"test": 9001}})
    assert admin_post.status_code == 201
    # Verify group created successfully
    verify_post = json.loads(admin_post.data)
    verify_group = Group.get_by_id(verify_post['id'])
    assert verify_post['vars'] == verify_group.vars
    # Test that creating a duplicate group fails
    dup_post = client.post(
        '/api/v1/groups/',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json={"name": "test group", "vars": {"different_var": 1138}})
    assert dup_post.status_code == 400


def test_groups_get(client, test_user, admin_user, test_inventory_bootstrap):
    # Test data
    test_group = Host.get_by_id_or_name('westprom').groups[0]
    # Create tokens
    create_token(test_user)
    create_token(admin_user)
    # Test that nopriv user cannot get a group
    nopriv_get = client.get(
        f'/api/v1/groups/{test_group.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert nopriv_get.status_code == 403
    # Test that getting an invalid group fails
    invalid_get = client.get(
        '/api/v1/groups/12345',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert invalid_get.status_code == 404
    # Test that admin can get a group
    admin_get = client.get(
        f'/api/v1/groups/{test_group.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert admin_get.status_code == 200
    admin_get_out = json.loads(admin_get.data)
    assert admin_get_out['name'] == test_group.name


def test_groups_patch(client, test_user, admin_user, test_inventory_bootstrap):
    # Test data
    test_group = Host.get_by_id_or_name('westprom').groups[0]
    test_data = {"vars": {"test": 1}}
    # Create tokens
    create_token(test_user)
    create_token(admin_user)
    # Test that nopriv user cannot patch a group
    nopriv_patch = client.patch(
        f'/api/v1/groups/{test_group.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)],
        json=test_data)
    assert nopriv_patch.status_code == 403
    # Test that patcing invalid group fails
    invalid_patch = client.patch(
        '/api/v1/groups/12345',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json=test_data)
    assert invalid_patch.status_code == 404
    # Test that admin can patch a group
    admin_patch = client.patch(
        f'/api/v1/groups/{test_group.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)],
        json=test_data)
    assert admin_patch.status_code == 200
    admin_patch_out = json.loads(admin_patch.data)
    assert admin_patch_out['vars'] == test_data['vars']


def test_groups_delete(client, test_user, admin_user, test_inventory_bootstrap):
    # Test data
    test_group = Host.get_by_id_or_name('westprom').groups[0]
    # Create tokens
    create_token(test_user)
    create_token(admin_user)
    # Test that nopriv user cannot delete a group
    nopriv_delete = client.delete(
        f'/api/v1/groups/{test_group.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', test_user.session_token)])
    assert nopriv_delete.status_code == 403
    # Test that deleting an invalid group fails
    invalid_delete = client.delete(
        '/api/v1/groups/12345',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert invalid_delete.status_code == 404
    # Test that admin can delete a group
    admin_delete = client.delete(
        f'/api/v1/groups/{test_group.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert admin_delete.status_code == 204
    verify = client.get(
        f'/api/v1/groups/{test_group.id}',
        follow_redirects=True,
        headers=[('X-AUTH-TOKEN', admin_user.session_token)])
    assert verify.status_code == 404



from functools import wraps
from flask import current_app, request
from flask_allows import allows, Or


def is_logged_in(user):
    """Checks if the user is logged in."""
    if current_app.config.get('LOGIN_DISABLED') or \
            (user.is_authenticated and user.is_active):
        return True
    return False


def is_same_user(user):
    """Checks if the user is the same user in the request."""
    user_name = request.view_args.get('user_name')
    return user.name == user_name


def has_permission(permission):

    def permission_checker(user):
        if current_app.config.get('LOGIN_DISABLED'):
            return True
        permissions = []
        # Everyone gets base permissions
        permissions.extend(current_app.config.opsy['auth']['base_permissions'])
        if user.is_authenticated and user.is_active:
            # Logged in users get logged in permissions
            permissions.extend(
                current_app.config.opsy['auth']['logged_in_permissions'])
            # And a user gets their own permissions from their roles
            if hasattr(user, 'permissions'):
                permissions.extend([x.name for x in user.permissions])
        return permission in permissions

    return permission_checker


def need_permission(permission_name, *requirements, identity=None,
                    on_fail=None, throws=None):
    """Modified version of flask_allows requires decorator to tie into RBAC."""

    def decorator(func):

        permission = has_permission(permission_name)
        if requirements:
            new_requirements = (Or(permission, *requirements),)
        else:
            new_requirements = (permission,)

        @wraps(func)
        def allower(*args, **kwargs):
            result = allows.run(
                new_requirements,
                identity=identity,
                on_fail=on_fail,
                throws=throws,
                f_args=args,
                f_kwargs=kwargs,
            )
            # authorization failed
            if result is not None:
                return result
            return func(*args, **kwargs)

        rbac_info = {
            'permission_needed': permission_name,
            'requirements': new_requirements
        }
        allower.__rbac__ = rbac_info
        return allower
    return decorator



import sys
import time
import logging
from logging.handlers import WatchedFileHandler
from flask import request, g
from flask_login import current_user


def log_before_request():
    """Needed to log on user logout since current_user is wiped."""
    if current_user.is_authenticated:
        g.logged_user = current_user.name
    else:
        g.logged_user = None


def log_after_request(response):
    """Apache style access logs."""
    access_logger = logging.getLogger('opsy_access')
    if current_user.is_authenticated:
        username = current_user.name
    else:
        username = g.logged_user or '-'
    timestamp = time.strftime("%d/%b/%Y:%H:%M:%S %z")
    length = response.headers.get('Content-Length') or '-'
    message = (
        f'{request.remote_addr} - {username} [{timestamp}] "{request.method} '
        f'{request.full_path} {request.environ.get("SERVER_PROTOCOL", "-")}" '
        f'{response.status_code} {length}')
    access_logger.info(message)
    return response


def configure_logging(config):
    # Nuke any default handlers
    logging.getLogger().handlers = []
    # First configure the application logger.
    app_logger = logging.getLogger('opsy')
    app_log_handlers = [logging.StreamHandler(sys.stdout)]
    if config['logging']['log_file']:
        app_log_handlers.append(
            WatchedFileHandler(config['logging']['log_file']))
    for log_handler in app_log_handlers:
        log_handler.setFormatter(
            logging.Formatter(
                '%(asctime)s %(process)d %(levelname)s %(module)s - '
                '%(message)s'))
        app_logger.addHandler(log_handler)
    app_logger.setLevel(config['logging']['log_level'])
    # Now configure the access logger.
    access_logger = logging.getLogger('opsy_access')
    access_log_handlers = [logging.StreamHandler(sys.stdout)]
    if config['logging']['access_log_file']:
        access_log_handlers.append(
            WatchedFileHandler(config['logging']['access_log_file']))
    for log_handler in access_log_handlers:
        log_handler.setFormatter(
            logging.Formatter('%(message)s'))
        access_logger.addHandler(log_handler)
    access_logger.setLevel(config['logging']['log_level'])



import asyncio
from abc import abstractmethod, ABCMeta
import aiohttp
from flask import current_app
from sqlalchemy import and_, tuple_
from werkzeug.urls import url_join
import opsy
from opsy.flask_extensions import db
from opsy.inventory.models import Host
from opsy.monitoring.models import Event
from opsy.monitoring.exceptions import PollFailure


class HttpPollerBackend(metaclass=ABCMeta):

    def __init__(self, monitoring_service, host='localhost', protocol='http',
                 port=80, path='/', interval=60, timeout=30, username=None,
                 password=None, verify_ssl=False):
        self.monitoring_service = monitoring_service
        self.host = host
        self.protocol = protocol
        self.port = port
        self.path = path
        self.interval = interval
        self.timeout = timeout
        self.username = username
        self.password = password
        self.verify_ssl = verify_ssl

    @property
    @abstractmethod
    def urls(self):
        """This should return a list of generated URLs to be polled."""

    @abstractmethod
    def decode(self, data):
        """This performs the data decoding. Returns list of raw events."""

    @property
    def base_url(self):
        return url_join(
            f'{self.protocol}://{self.host}:{self.port}', self.path)

    @property
    def config(self):
        """This returns the config for the poller as a dict."""
        return {
            'host': self.host,
            'protocol': self.protocol,
            'port': self.port,
            'path': self.path,
            'interval': self.interval,
            'timeout': self.timeout,
            'username': self.username,
            'password': self.password,
            'verify_ssl': self.verify_ssl
        }

    def _create_session(self):
        auth = aiohttp.BasicAuth(self.username, self.password) \
            if (self.username and self.password) else None
        conn = aiohttp.TCPConnector(verify_ssl=False) \
            if not self.verify_ssl else None
        headers = {'User-Agent': 'Opsy/%s' % opsy.__version__}
        timeout = aiohttp.ClientTimeout(self.timeout)
        return aiohttp.ClientSession(auth=auth, connector=conn,
                                     headers=headers, timeout=timeout)

    async def fetch(self, urls, expected_status=None):
        """This performs the data retrival."""
        async with self._create_session() as session:
            tasks = [
                asyncio.ensure_future(
                    self.get(session, url, expected_status=expected_status))
                for url in urls]
            return await asyncio.gather(*tasks)

    def update(self):
        """This triggers the work needed to update."""
        current_app.logger.info(
            f'Updating monitoring service {self.monitoring_service.name}..')
        try:
            loop = asyncio.get_event_loop()
            data = loop.run_until_complete(self.fetch(self.urls))
        except PollFailure as error:
            current_app.logger.error(  # pylint: disable=no-member
                'Failed to update events on monitoring service %s: %s',
                self.monitoring_service.name, error.__class__.__name__)

        # So basic idea of what we're doing in this next section:
        # 1) We get the list of all events from the API.
        new_events = self.decode(data)
        # 2) We create a key tuple of all the events in a list.
        event_keys = [(self.monitoring_service.id,
                       x['host_name'],
                       x['check_name']) for x in new_events]
        # 3) We get a list of all the hosts with names matching the events.
        hosts = Host.query.filter(
            Host.name.in_([x['host_name'] for x in new_events])).all()
        # 4) We compare this to all unresolved events to find the ones that
        #    resolved since we last polled and mark them as such.
        for resolved_event in Event.query.filter(and_(
                Event.monitoring_service_id == self.monitoring_service.id,
                ~Event.resolved,
                ~tuple_(
                    Event.monitoring_service_id,
                    Event.host_name,
                    Event.check_name).in_(event_keys))).all():
            resolved_event.resolve(commit=False)
            db.session.add(resolved_event)
        # 5) We compare this to all the events we currently know about and mark
        #    those for update so we don't get duplicates. We also add the
        #    host_id if the host exists.
        existing_events = []
        for existing_event in Event.query.filter(and_(
                Event.monitoring_service_id == self.monitoring_service.id,
                ~Event.resolved,
                tuple_(
                    Event.monitoring_service_id,
                    Event.host_name,
                    Event.check_name).in_(event_keys))).all():
            for new_event in new_events:
                if new_event['host_name'] == existing_event.host_name and \
                        new_event['check_name'] == existing_event.check_name:
                    host_id = None
                    for host in hosts:
                        if new_event['host_name'] == host.name:
                            host_id = host.id
                            break
                    existing_events.append(new_event)
                    db.session.add(
                        existing_event.update(commit=False, host_id=host_id,
                                              **new_event))
                    break
        # 6) We eliminate the existing events from the new list.
        new_events = [x for x in new_events if x not in existing_events]
        # 7) We create the new events, adding the host_id if the host exists.
        for new_event in new_events:
            current_app.logger.debug(f'Adding new event: {new_event}')
            host_id = None
            for host in hosts:
                if new_event['host_name'] == host.name:
                    host_id = host.id
                    break
            db.session.add(
                Event(self.monitoring_service, host_id=host_id, **new_event))
        # 8) And commit the changes.
        db.session.commit()
        current_app.logger.info(
            f'Updated monitoring service {self.monitoring_service.name}.')

    async def get(self, session, url, expected_status=None):
        expected_status = expected_status or [200]
        try:
            response = await session.get(url)
        except asyncio.TimeoutError:
            raise aiohttp.ClientError('Timeout exceeded')
        if response.status not in expected_status:
            response.close()
            raise aiohttp.ClientError('Unexpected response from %s, got'
                                      ' %s' % (url, response.status))
        return await response.json()



import os.path
import toml
from marshmallow import (Schema, fields, validate, validates_schema,
                         ValidationError)
from opsy.exceptions import NoConfigFile


class ConfigAppSchema(Schema):
    database_uri = fields.Str(required=True)
    secret_key = fields.Str(required=True)
    uri_prefix = fields.Str(missing='/')


class ConfigAuthSchema(Schema):
    base_permissions = fields.List(fields.Str(), missing=[])
    logged_in_permissions = fields.List(fields.Str(), missing=[])
    session_token_ttl = fields.Integer(missing=86400)
    ldap_enabled = fields.Boolean(missing=False)
    ldap_host = fields.Str(missing=None)
    ldap_port = fields.Integer(missing=389)
    ldap_use_ssl = fields.Boolean(missing=False)
    ldap_bind_user_dn = fields.Str(missing=None)
    ldap_bind_user_password = fields.Str(missing=None)
    ldap_base_dn = fields.Str(missing='')
    ldap_user_dn = fields.Str(missing='')
    ldap_user_object_filter = fields.Str(missing='(objectclass=person)')
    ldap_user_login_attr = fields.Str(missing='uid')
    ldap_user_rdn_attr = fields.Str(missing='uid')
    ldap_user_full_name_attr = fields.Str(missing='displayName')
    ldap_user_email_attr = fields.Str(missing='mail')
    ldap_user_search_scope = fields.Str(
        validate=validate.OneOf(['LEVEL', 'SUBTREE']), missing='LEVEL')
    ldap_group_dn = fields.Str(missing='')
    ldap_group_object_filter = fields.Str(missing='(objectclass=groupOfNames)')
    ldap_group_members_attr = fields.Str(missing='Member')
    ldap_group_name_attr = fields.Str(missing='cn')
    ldap_group_search_scope = fields.Str(
        validate=validate.OneOf(['LEVEL', 'SUBTREE']), missing='LEVEL')

    @validates_schema
    def validate_ldap(self, data, **kwargs):
        if data.get('ldap_enabled') and not data.get('ldap_host'):
            raise ValidationError(
                'Auth option "ldap_host" is required when "ldap_enabled" is '
                'set to true.')


class ConfigLoggingSchema(Schema):
    log_file = fields.Str(missing=None)
    access_log_file = fields.Str(missing=None)
    log_level = fields.Str(
        validate=validate.OneOf(
            ['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG']),
        missing='INFO')


class ConfigServerSchema(Schema):
    host = fields.Str(missing='localhost')
    port = fields.Integer(missing=5000)
    threads = fields.Integer(missing=10)
    ssl_enabled = fields.Boolean(missing=False)
    certificate = fields.Str(missing=None)
    private_key = fields.Str(missing=None)
    ca_certificate = fields.Str(missing=None)

    @validates_schema
    def validate_ssl(self, data, **kwargs):
        if data.get('ssl_enabled'):
            certificate = data.get('certificate')
            private_key = data.get('private_key')
            if not all([certificate, private_key]):
                raise ValidationError(
                    'Server options "certificate" and "private_key" are '
                    'required when "ssl_enabled" is set to true.')


class ConfigSchema(Schema):
    app = fields.Nested(ConfigAppSchema(), required=True)
    auth = fields.Nested(
        ConfigAuthSchema(), missing=ConfigAuthSchema().load({}))
    logging = fields.Nested(
        ConfigLoggingSchema(), missing=ConfigLoggingSchema().load({}))
    server = fields.Nested(
        ConfigServerSchema(), missing=ConfigServerSchema().load({}))


def load_config(config_file):
    if not os.path.exists(config_file):
        raise NoConfigFile(f'File "{config_file}" does not exist.')
    with open(config_file, 'r') as file_handler:
        config = ConfigSchema().load(toml.load(file_handler))
    return config


def configure_app(app, config):
    app.config.update({
        'SQLALCHEMY_TRACK_MODIFICATIONS': False,
        'JSON_SORT_KEYS': False,
        'SECRET_KEY': config['app']['secret_key'],
        'SQLALCHEMY_DATABASE_URI': config['app']['database_uri'],
        'LDAP_HOST': config['auth']['ldap_host'],
        'LDAP_PORT': config['auth']['ldap_port'],
        'LDAP_USE_SSL': config['auth']['ldap_use_ssl'],
        'LDAP_BIND_USER_DN': config['auth']['ldap_bind_user_dn'],
        'LDAP_BIND_USER_PASSWORD': config['auth']['ldap_bind_user_password'],
        'LDAP_BASE_DN': config['auth']['ldap_base_dn'],
        'LDAP_USER_DN': config['auth']['ldap_user_dn'],
        'LDAP_USER_OBJECT_FILTER': config['auth']['ldap_user_object_filter'],
        'LDAP_USER_LOGIN_ATTR': config['auth']['ldap_user_login_attr'],
        'LDAP_USER_RDN_ATTR': config['auth']['ldap_user_rdn_attr'],
        'LDAP_USER_SEARCH_SCOPE': config['auth']['ldap_user_search_scope'],
        'LDAP_GROUP_DN': config['auth']['ldap_group_dn'],
        'LDAP_GROUP_OBJECT_FILTER': config['auth']['ldap_group_object_filter'],
        'LDAP_GROUP_MEMBERS_ATTR': config['auth']['ldap_group_members_attr'],
        'LDAP_GROUP_SEARCH_SCOPE': config['auth']['ldap_group_search_scope']
    })
    setattr(app.config, 'opsy', config)



import pytest
from collections import OrderedDict
from opsy.auth.schema import (
    UserLoginSchema, AppPermissionSchema, UserTokenSchema, UserSchema,
    RoleSchema, RolePermissionSchema)
from opsy.auth.utils import create_token
from opsy.auth.models import Permission
from marshmallow.exceptions import ValidationError


###############################################################################
# UserLoginSchema Tests
###############################################################################


def test_user_login_schema():

    def make_payload(u='user', p='pass', f=False):
        """Return schema payload with all arguments."""
        return {'user_name': u, 'password': p, 'force_renew': f}

    payload = make_payload()
    expected_user_login_schema_input = payload
    test_user = payload
    assert UserLoginSchema().load(test_user) == expected_user_login_schema_input

    nouser = make_payload(u=None)
    with pytest.raises(ValidationError):
        UserLoginSchema().load(nouser)

    nopass = make_payload(p=None)
    with pytest.raises(ValidationError):
        UserLoginSchema().load(nopass)


###############################################################################
# AppPermissionSchema Tests
###############################################################################


def test_app_permission_schema():

    expected_permission_schema_input = {
        'endpoint': 'test',
        'method': 'test',
        'permission_needed': 'test'}
    expected_permission_schema_output = OrderedDict([
        ('endpoint', 'test'),
        ('method', 'test'),
        ('permission_needed', 'test')])

    assert AppPermissionSchema().dump(expected_permission_schema_input) \
        == expected_permission_schema_output


###############################################################################
# UserSchema Tests
###############################################################################


def test_user_schema(test_user):

    expected_user_schema_output = OrderedDict([
        ('id', test_user.id),
        ('name', test_user.name),
        ('full_name', test_user.full_name),
        ('email', test_user.email),
        ('enabled', test_user.enabled),
        ('ldap_user', test_user.ldap_user),
        ('created_at', test_user.created_at.isoformat()),
        ('updated_at', test_user.updated_at.isoformat()),
        ('roles', [
            OrderedDict([
                ('id', x.id),
                ('name', x.name),
                ('_links', {
                    'self': f'/api/v1/roles/{x.id}',
                    'collection': '/api/v1/roles/'})])
            for x in test_user.roles]),
        ('permissions', [OrderedDict([
            ('id', x.id),
            ('name', x.name),
            ('_links', {
                'self': f'/api/v1/roles/{x.role.id}/permissions/{x.id}',
                'collection': f'/api/v1/roles/{x.role.id}/permissions/'})])
            for x in test_user.permissions]),
        ('_links', {
            'self': f'/api/v1/users/{test_user.id}',
            'collection': '/api/v1/users/'})])

    assert UserSchema().dump(test_user) == expected_user_schema_output


###############################################################################
# UserTokenSchema Tests
###############################################################################


def test_user_token_schema(test_user):

    create_token(test_user)

    expected_zone_schema_output = OrderedDict([
        ('user_id', test_user.id),
        ('user_name', test_user.name),
        ('full_name', test_user.full_name),
        ('email', test_user.email),
        ('ldap_user', test_user.ldap_user),
        ('token', test_user.session_token),
        ('expires_at', test_user.session_token_expires_at.isoformat())])

    assert UserTokenSchema().dump(test_user) == expected_zone_schema_output


###############################################################################
# RoleSchema Tests
###############################################################################


def test_role_schema(test_role):

    expected_role_schema_output = OrderedDict([
        ('id', test_role.id),
        ('name', test_role.name),
        ('ldap_group', test_role.ldap_group),
        ('description', test_role.description),
        ('created_at', test_role.created_at.isoformat()),
        ('updated_at', test_role.updated_at.isoformat()),
        ('permissions', [OrderedDict([
            ('id', x.id),
            ('name', x.name),
            ('_links', {
                'self': f'/api/v1/roles/{x.role.id}/permissions/{x.id}',
                'collection': f'/api/v1/roles/{x.role.id}/permissions/'})])
            for x in test_role.permissions]),
        ('users', test_role.users),
        ('_links', {
            'self': f'/api/v1/roles/{test_role.id}',
            'collection': '/api/v1/roles/'})])

    assert RoleSchema().dump(test_role) == expected_role_schema_output


###############################################################################
# RolePermissionSchema Tests
###############################################################################


def test_role_permission_schema(admin_user):

    test_perm = Permission.query.first()

    expected_role_permission_schema_output = OrderedDict([
        ('id', test_perm.id),
        ('role', OrderedDict([
            ('id', test_perm.role.id),
            ('name', test_perm.role.name),
            ('_links', {
                'self': f'/api/v1/roles/{test_perm.role.id}',
                'collection': '/api/v1/roles/'})])),
        ('name', test_perm.name),
        ('created_at', test_perm.created_at.isoformat()),
        ('updated_at', test_perm.updated_at.isoformat()),
        ('_links', {
            'self': f'/api/v1/roles/{test_perm.role.id}/permissions/{test_perm.id}',
            'collection': f'/api/v1/roles/{test_perm.role.id}/permissions/'})])

    assert RolePermissionSchema().dump(test_perm) \
        == expected_role_permission_schema_output



from flask import abort, Blueprint
from flask_apispec import marshal_with, doc
from opsy.flask_extensions import apispec
from opsy.rbac import need_permission
from opsy.schema import use_kwargs, EmptySchema
from opsy.inventory.schema import (
    ZoneSchema, ZoneUpdateSchema, ZoneQuerySchema,
    HostSchema, HostUpdateSchema, HostQuerySchema,
    GroupSchema, GroupUpdateSchema, GroupQuerySchema,
    HostGroupMappingSchema, HostGroupMappingUpdateSchema,
    HostGroupMappingQuerySchema)
from opsy.inventory.models import Zone, Host, Group, HostGroupMapping
from opsy.exceptions import DuplicateError


def create_inventory_views(app):
    app.register_blueprint(zones_blueprint, url_prefix='/api/v1/zones')
    app.register_blueprint(hosts_blueprint, url_prefix='/api/v1/hosts')
    app.register_blueprint(groups_blueprint, url_prefix='/api/v1/groups')
    apispec.spec.tag(
        {'name': 'zones',
         'description': 'Zones are the base grouping for inventory in Opsy.'})
    for view in [zones_list, zones_post, zones_get, zones_patch,
                 zones_delete]:
        apispec.register(view, blueprint='inventory_zones')
    apispec.spec.tag(
        {'name': 'hosts',
         'description': 'Hosts are your servers.'})
    for view in [hosts_list, hosts_post, hosts_get, hosts_patch, hosts_delete,
                 host_group_mappings_list, host_group_mappings_post,
                 host_group_mappings_get, host_group_mappings_patch,
                 host_group_mappings_delete]:
        apispec.register(view, blueprint='inventory_hosts')
    apispec.spec.tag(
        {'name': 'groups',
         'description': 'Groups are logical groupings of your hosts.'})
    for view in [groups_list, groups_post, groups_get, groups_patch,
                 groups_delete]:
        apispec.register(view, blueprint='inventory_groups')

###############################################################################
# Blueprints
###############################################################################


# pylint: disable=invalid-name
zones_blueprint = Blueprint('inventory_zones', __name__)
# pylint: disable=invalid-name
hosts_blueprint = Blueprint('inventory_hosts', __name__)
# pylint: disable=invalid-name
groups_blueprint = Blueprint('inventory_groups', __name__)

###############################################################################
# Zone Views
###############################################################################


@zones_blueprint.route('/', methods=['GET'])
@use_kwargs(ZoneQuerySchema, locations=['query'])
@marshal_with(ZoneSchema(many=True), code=200)
@doc(
    operationId='list_zones',
    summary='List zones.',
    description='',
    tags=['zones'],
    security=[{'api_key': []}])
@need_permission('list_zones')
def zones_list(**kwargs):
    return Zone.query.filter_in(**kwargs).all()


@zones_blueprint.route('/', methods=['POST'])
@use_kwargs(ZoneSchema)
@marshal_with(ZoneSchema(), code=201)
@doc(
    operationId='create_zone',
    summary='Create a new zone.',
    description='',
    tags=['zones'],
    security=[{'api_key': []}])
@need_permission('create_zone')
def zones_post(**kwargs):
    try:
        return Zone.create(**kwargs), 201
    except (DuplicateError, ValueError) as error:
        abort(400, str(error))


@zones_blueprint.route('/<id_or_name>', methods=['GET'])
@marshal_with(ZoneSchema(), code=200)
@doc(
    operationId='show_zone',
    summary='Show a zone.',
    description='',
    tags=['zones'],
    security=[{'api_key': []}])
@need_permission('show_zone')
def zones_get(id_or_name):
    try:
        return Zone.get_by_id_or_name(id_or_name)
    except ValueError as error:
        abort(404, str(error))


@zones_blueprint.route('/<id_or_name>', methods=['PATCH'])
@use_kwargs(ZoneUpdateSchema)
@marshal_with(ZoneSchema(), code=200)
@doc(
    operationId='update_zone',
    summary='Update a zone.',
    description='',
    tags=['zones'],
    security=[{'api_key': []}])
@need_permission('update_zone')
def zones_patch(id_or_name, **kwargs):
    try:
        return Zone.update_by_id_or_name(id_or_name, **kwargs)
    except ValueError as error:
        abort(404, str(error))


@zones_blueprint.route('/<id_or_name>', methods=['DELETE'])
@marshal_with(EmptySchema, code=204)
@doc(
    operationId='delete_zone',
    summary='Delete a zone.',
    description='',
    tags=['zones'],
    security=[{'api_key': []}])
@need_permission('delete_zone')
def zones_delete(id_or_name):
    try:
        Zone.delete_by_id_or_name(id_or_name)
        return ('', 204)
    except ValueError as error:
        abort(404, str(error))

###############################################################################
# Host Views
###############################################################################


@hosts_blueprint.route('/', methods=['GET'])
@use_kwargs(HostQuerySchema, locations=['query'])
@marshal_with(HostSchema(many=True), code=200)
@doc(
    operationId='list_hosts',
    summary='List hosts.',
    description='',
    tags=['hosts'],
    security=[{'api_key': []}])
@need_permission('list_hosts')
def hosts_list(**kwargs):
    return Host.query.filter_in(**kwargs).all()


@hosts_blueprint.route('/', methods=['POST'])
@use_kwargs(HostSchema)
@marshal_with(HostSchema(), code=201)
@doc(
    operationId='create_host',
    summary='Create a new host.',
    description='',
    tags=['hosts'],
    security=[{'api_key': []}])
@need_permission('create_host')
def hosts_post(**kwargs):
    try:
        return Host.create(**kwargs), 201
    except (DuplicateError, ValueError) as error:
        abort(400, str(error))


@hosts_blueprint.route('/<id_or_name>', methods=['GET'])
@marshal_with(HostSchema(), code=200)
@doc(
    operationId='show_host',
    summary='Show a host.',
    description='',
    tags=['hosts'],
    security=[{'api_key': []}])
@need_permission('show_host')
def hosts_get(id_or_name):
    try:
        return Host.get_by_id_or_name(id_or_name)
    except ValueError as error:
        abort(404, str(error))


@hosts_blueprint.route('/<id_or_name>', methods=['PATCH'])
@use_kwargs(HostUpdateSchema)
@marshal_with(HostSchema(), code=200)
@doc(
    operationId='update_host',
    summary='Update a host.',
    description='',
    tags=['hosts'],
    security=[{'api_key': []}])
@need_permission('update_host')
def hosts_patch(id_or_name, **kwargs):
    try:
        return Host.update_by_id_or_name(id_or_name, **kwargs)
    except ValueError as error:
        abort(404, str(error))


@hosts_blueprint.route('/<id_or_name>', methods=['DELETE'])
@marshal_with(EmptySchema, code=204)
@doc(
    operationId='delete_host',
    summary='Delete a host.',
    description='',
    tags=['hosts'],
    security=[{'api_key': []}])
@need_permission('delete_host')
def hosts_delete(id_or_name):
    try:
        Host.delete_by_id_or_name(id_or_name)
        return ('', 204)
    except ValueError as error:
        abort(404, str(error))


###############################################################################
# Host Group Mapping Views
###############################################################################


@hosts_blueprint.route('/<id_or_name>/group_mappings/', methods=['GET'])
@use_kwargs(HostGroupMappingQuerySchema, locations=['query'])
@marshal_with(HostGroupMappingSchema(many=True), code=200)
@doc(
    operationId='list_group_mappings',
    summary='List group mappings.',
    description='',
    tags=['hosts'],
    security=[{'api_key': []}])
@need_permission('show_host')
def host_group_mappings_list(id_or_name, **kwargs):
    try:
        host = Host.get_by_id_or_name(id_or_name)
    except ValueError as error:
        abort(404, str(error))
    return HostGroupMapping.query.filter_by(
        host_id=host.id).filter_in(**kwargs).all()


@hosts_blueprint.route('/<id_or_name>/group_mappings/', methods=['POST'])
@use_kwargs(HostGroupMappingSchema)
@marshal_with(HostGroupMappingSchema(), code=201)
@doc(
    operationId='create_group_mapping',
    summary='Create a group mapping.',
    description='',
    tags=['hosts'],
    security=[{'api_key': []}])
@need_permission('update_host')
def host_group_mappings_post(id_or_name, **kwargs):
    try:
        host = Host.get_by_id_or_name(id_or_name)
        group = Group.get_by_id(kwargs['group_id'])
    except ValueError as error:
        abort(404, str(error))
    try:
        return host.add_group(group, priority=kwargs.get('priority')), 201
    except (DuplicateError, ValueError) as error:
        abort(400, str(error))


@hosts_blueprint.route(
    '/<id_or_name>/group_mappings/<group_id_or_name>',
    methods=['GET'])
@marshal_with(HostGroupMappingSchema(), code=200)
@doc(
    operationId='show_group_mapping',
    summary='Show a group mapping.',
    description='',
    tags=['hosts'],
    security=[{'api_key': []}])
@need_permission('show_host')
def host_group_mappings_get(id_or_name, group_id_or_name):
    try:
        return HostGroupMapping.get_by_host_and_group(
            id_or_name, group_id_or_name)
    except ValueError as error:
        abort(404, str(error))


@hosts_blueprint.route(
    '/<id_or_name>/group_mappings/<group_id_or_name>',
    methods=['PATCH'])
@use_kwargs(HostGroupMappingUpdateSchema)
@marshal_with(HostGroupMappingSchema(), code=200)
@doc(
    operationId='update_group_mapping',
    summary='Update a group mapping.',
    description='',
    tags=['hosts'],
    security=[{'api_key': []}])
@need_permission('update_host')
def host_group_mappings_patch(id_or_name, group_id_or_name, **kwargs):
    try:
        return HostGroupMapping.get_by_host_and_group(
            id_or_name, group_id_or_name).update(**kwargs)
    except ValueError as error:
        abort(404, str(error))


@hosts_blueprint.route(
    '/<id_or_name>/group_mappings/<group_id_or_name>',
    methods=['DELETE'])
@marshal_with(EmptySchema, code=204)
@doc(
    operationId='delete_group_mapping',
    summary='Delete a group mapping.',
    description='',
    tags=['hosts'],
    security=[{'api_key': []}])
@need_permission('update_host')
def host_group_mappings_delete(id_or_name, group_id_or_name):
    try:
        HostGroupMapping.get_by_host_and_group(
            id_or_name, group_id_or_name).delete()
        return ('', 204)
    except ValueError as error:
        abort(404, str(error))


###############################################################################
# Group Views
###############################################################################


@groups_blueprint.route('/', methods=['GET'])
@use_kwargs(GroupQuerySchema, locations=['query'])
@marshal_with(GroupSchema(many=True), code=200)
@doc(
    operationId='list_groups',
    summary='List groups.',
    description='',
    tags=['groups'],
    security=[{'api_key': []}])
@need_permission('list_groups')
def groups_list(**kwargs):
    return Group.query.filter_in(**kwargs).all()


@groups_blueprint.route('/', methods=['POST'])
@use_kwargs(GroupSchema)
@marshal_with(GroupSchema(), code=201)
@doc(
    operationId='create_group',
    summary='Create a group.',
    description='',
    tags=['groups'],
    security=[{'api_key': []}])
@need_permission('create_group')
def groups_post(**kwargs):
    try:
        return Group.create(**kwargs), 201
    except (DuplicateError, ValueError) as error:
        abort(400, str(error))


@groups_blueprint.route('/<group_id>', methods=['GET'])
@marshal_with(GroupSchema(), code=200)
@doc(
    operationId='show_group',
    summary='Show a group.',
    description='',
    tags=['groups'],
    security=[{'api_key': []}])
@need_permission('show_group')
def groups_get(group_id):
    try:
        return Group.get_by_id(group_id)
    except ValueError as error:
        abort(404, str(error))


@groups_blueprint.route('/<group_id>', methods=['PATCH'])
@use_kwargs(GroupUpdateSchema)
@marshal_with(GroupSchema(), code=200)
@doc(
    operationId='update_group',
    summary='Update a group.',
    description='',
    tags=['groups'],
    security=[{'api_key': []}])
@need_permission('update_group')
def groups_patch(group_id, **kwargs):
    try:
        return Group.update_by_id(group_id, **kwargs)
    except ValueError as error:
        abort(404, str(error))


@groups_blueprint.route('/<group_id>', methods=['DELETE'])
@marshal_with(EmptySchema, code=204)
@doc(
    operationId='delete_group',
    summary='Delete a group.',
    description='',
    tags=['groups'],
    security=[{'api_key': []}])
@need_permission('delete_group')
def groups_delete(group_id):
    try:
        Group.delete_by_id(group_id)
        return ('', 204)
    except ValueError as error:
        abort(404, str(error))
